package edu.kit.ipd.descartes.adaptation.model.adaptation.dmm;

import java.io.IOException;
import java.util.List;
import java.util.Random;

import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;

import edu.kit.ipd.descartes.adaptation.model.adaptation.IActionHandler;
import edu.kit.ipd.descartes.adaptation.model.adaptation.dmm.util.DmmModelActionHelper;
import edu.kit.ipd.descartes.adaptation.model.adaptation.dmm.util.DmmModelChanger;
import edu.kit.ipd.descartes.adaptation.model.adaptation.dmm.util.OclEvaluationHelper;
import edu.kit.ipd.descartes.adaptation.model.adaptation.exceptions.OperationNotPerformedException;
import edu.kit.ipd.descartes.adaptation.model.repository.dmm.DmlModelLoader;
import edu.kit.ipd.descartes.core.AdaptableEntity;
import edu.kit.ipd.descartes.core.Entity;
import edu.kit.ipd.descartes.mm.adaptation.AdaptationActionOperation;
import edu.kit.ipd.descartes.mm.adaptation.AdaptationScope;
import edu.kit.ipd.descartes.mm.adaptationpoints.AdaptationPoint;
import edu.kit.ipd.descartes.mm.adaptationpoints.ModelEntityConfigurationRange;
import edu.kit.ipd.descartes.mm.adaptationpoints.ModelVariableConfigurationRange;
import edu.kit.ipd.descartes.mm.adaptationpoints.PropertyRange;
import edu.kit.ipd.descartes.mm.adaptationpoints.SetOfConfigurations;
import edu.kit.ipd.descartes.mm.adaptationpoints.VariationType;
import edu.kit.ipd.descartes.mm.containerrepository.ContainerTemplate;
import edu.kit.ipd.descartes.mm.containerrepository.ContainerrepositoryPackage;
import edu.kit.ipd.descartes.mm.resourcelandscape.Container;
import edu.kit.ipd.descartes.mm.resourcelandscape.ResourcelandscapePackage;
import edu.kit.ipd.descartes.mm.resourcelandscape.RuntimeEnvironment;

public class DmmModelActionHandler implements IActionHandler {
    static Logger logger = Logger.getLogger(DmmModelActionHandler.class);

    public DmmModelActionHandler() {
        logger.info("DMM models loaded.");
    }

    private void saveModels() {
        try {
            DmlModelLoader.saveAll();
        } catch (IOException e) {
            logger.error("Error while saving models", e);
        }
    }

    @Override
    public void execute(AdaptationPoint point, AdaptationActionOperation operation)
            throws OperationNotPerformedException {

        logger.debug("Executing adaptation action <<" + operation.getAdaptationOperationDirection().getLiteral() + ", "
                + operation.getAdaptationOperationScope().getLiteral() + ">> on adaptation point <<" + point.getName()
                + ">>.");

        if (point instanceof ModelEntityConfigurationRange) {
            logger.debug("Type of Adaptation: EntityConfigurationRange");
            execEntityConfigRangeAdaptation(point, operation);
        } else if (point instanceof ModelVariableConfigurationRange) {
            logger.debug("Type of Adaptation Point: VariableConfigurationRange");
            execVariableConfigRangeAdaptation(point, operation);
        } else {
            logger.error("No matching type found for given adaptation point " + point.toString());
            throw new OperationNotPerformedException(
                    "No maching adaptation point execution found for adaptation point " + point);
        }
    }

    public void execVariableConfigRangeAdaptation(AdaptationPoint point, AdaptationActionOperation operation)
            throws OperationNotPerformedException {
        ModelVariableConfigurationRange configRange = (ModelVariableConfigurationRange) point;

        AdaptableEntity adaptedEntity = configRange.getAdaptableEntity();
        EObject actuallyConfiguredInstance = DmmModelActionHelper.findParentConfiguredInstance(adaptedEntity);
        switch (actuallyConfiguredInstance.eClass().getClassifierID()) {
        case ResourcelandscapePackage.RUNTIME_ENVIRONMENT:
            DmmModelChanger.scaleModelVariable(configRange, operation.getAdaptationOperationDirection());
            break;
        case ContainerrepositoryPackage.CONTAINER_TEMPLATE:
            DmmModelChanger.scaleContainerTemplate((ContainerTemplate) actuallyConfiguredInstance, configRange,
                    operation);
            break;
        case ResourcelandscapePackage.CONTAINER:
            /*
             * This code should not be reached because RUNTIME_ENVIRONMENTS are also containers.
             * Only AbstractHardwareInfrastructures are Containers but not RuntimeEnvironments, but
             * they cannot be changed because they're hardware
             */
            logger.error("You cannot specify an adaptation point for " + actuallyConfiguredInstance.eClass());
            break;
        default:
            logger.error("No adaptation action implemented for type " + actuallyConfiguredInstance.eClass());
            break;
        }
    }

    public void execEntityConfigRangeAdaptation(AdaptationPoint point, AdaptationActionOperation operation)
            throws OperationNotPerformedException {

        ModelEntityConfigurationRange configRange = (ModelEntityConfigurationRange) point;
        Entity adaptableEntity = configRange.getAdaptableEntity();
        VariationType range = configRange.getVariationPossibility();

        switch (operation.getAdaptationOperationDirection()) {
        case INCREASE:
            scaleOutModelEntity(adaptableEntity, range);
            break;
        case DECREASE:
            scaleInModelEntity(adaptableEntity, range);
            break;
        case MIGRATE:
            logger.debug("Executing MIGRATE operation");
            migrate(adaptableEntity, range, operation.getAdaptationOperationScope());
            break;
        default:
            throw new OperationNotPerformedException("Operation " + operation + " on adaptation point "
                    + point.getName() + " not supported");
        }
    }

    private void scaleInModelEntity(Entity adaptableEntity, VariationType range) throws OperationNotPerformedException {
        logger.debug("Executing DECREASE operation");
        Entity scaledEntity = scaleIn(adaptableEntity);
        if (scaledEntity == null)
            throw new OperationNotPerformedException("Could not scale " + adaptableEntity.getName()
                    + ". Minimum has been reached!");
    }

    private void scaleOutModelEntity(Entity adaptableEntity, VariationType range) throws OperationNotPerformedException {
        logger.debug("Executing INCREASE operation");
        RuntimeEnvironment scaledRE = scaleOut(adaptableEntity);
        if (!invariantsAreValid(adaptableEntity, (PropertyRange) range)) { // spawn new
                                                                           // runtimeEnvironment on
                                                                           // different machine
            List<Container> possibleTargets = DmmModelActionHelper.queryContainersWithSimilarTemplate(scaledRE
                    .getContainedIn());
            for (Container nextContainer : possibleTargets) {
                scaledRE.setContainedIn(nextContainer);
                if (invariantsAreValid(scaledRE, (PropertyRange) range)) {
                    logger.debug("Scaling entity " + scaledRE.getName() + " on altenative container "
                            + nextContainer.getName() + " (ID: " + nextContainer.getId() + ")");
                    return;
                }
            }
            // in case a violation occured and no alternative container was found
            EcoreUtil.delete(scaledRE, true);
            throw new OperationNotPerformedException("Could not scale " + adaptableEntity.getName()
                    + ". Maxiumum has been reached!");
        }
    }

    private boolean invariantsAreValid(Entity adaptableEntity, PropertyRange range) {
        boolean invariant = OclEvaluationHelper.evaluateOclConstraint(adaptableEntity, range.getMinValueConstraint()
                .getOclString(), adaptableEntity.eClass());

        if (!invariant)
            logger.debug("Invariant " + range.getMinValueConstraint().getOclString() + " violated");

        invariant &= OclEvaluationHelper.evaluateOclConstraint(adaptableEntity, range.getMaxValueConstraint()
                .getOclString(), adaptableEntity.eClass());

        if (!invariant)
            logger.debug("Invariant " + range.getMaxValueConstraint().getOclString() + " violated");

        return invariant;
    }

    private void migrate(Entity adaptableEntity, VariationType targets, AdaptationScope adaptationScope) throws OperationNotPerformedException {
        switch (adaptableEntity.eClass().getClassifierID()) {
        case ResourcelandscapePackage.RUNTIME_ENVIRONMENT:
            logger.debug("Starting migration of a runtime environment referring to container template "
                    + adaptableEntity.getName() + " (ID: " + adaptableEntity.getId() + ")");
            /*
             * 1. Copy RE 2. Attach copy to a target specified in the Adaptation Point 3. Delete
             * origin
             */

            if (!(targets instanceof SetOfConfigurations))
                throw new IllegalArgumentException(
                        "Migration could not be performed because no target specification found!");
            EList<Entity> migrationTargets = ((SetOfConfigurations) targets).getVariants();

            RuntimeEnvironment migratedEntity = (RuntimeEnvironment) adaptableEntity;
            Container source = migratedEntity.getContainedIn();
            // TODO Here you could implement something more sophisticated
            // e.g., choose the container with the lowest util.
            Entity target = randomTarget(migrationTargets, source);
            
            if (!(target instanceof Container))
                throw new IllegalArgumentException(
                        "Migration could not be performed because target is not a container!");
            Container targetContainer = (Container) target;
            
            if (migrationTargets.size() == 1 && EcoreUtil.equals(source, target))
                throw new OperationNotPerformedException("Migration operation not executed because source and destination are identical.");

            targetContainer.getContains().add(migratedEntity);
            source.getContains().remove(migratedEntity);
            break;
        case ContainerrepositoryPackage.CONTAINER_TEMPLATE:
//            logger.debug("Starting migration of a container referring to container template "
//                    + adaptableEntity.getName() + " (ID: " + adaptableEntity.getId() + ")");
            ContainerTemplate containerTemplate = (ContainerTemplate) adaptableEntity;
            Container migratedContainer = DmmModelActionHelper.findSuitableContainer(containerTemplate, adaptationScope, targets);
            if (migratedContainer == null)
                throw new OperationNotPerformedException("Could not execute migration operation because no suitable migration candidate found");
            migrate(migratedContainer, targets, adaptationScope);
            break;
        default:
            throw new IllegalArgumentException("No implemented migrate operation for " + adaptableEntity);
        }

    }

    private Entity randomTarget(EList<Entity> destinations, Entity source) {
        if (destinations.size() == 1)
            return destinations.get(0);
        
        Entity destination = destinations.get(0);
        Random randomGenerator = new Random();
        while (EcoreUtil.equals(destination, source)) {
            int destNumber = randomGenerator.nextInt(destinations.size());
            destination = destinations.get(destNumber);
        }

        return destination;
    }

    private Entity scaleIn(Entity adaptableEntity) {
        Entity result = null;
        switch (adaptableEntity.eClass().getClassifierID()) {
        case ResourcelandscapePackage.RUNTIME_ENVIRONMENT:
            result = DmmModelChanger.scaleInRuntimeEnvironmentInstance((RuntimeEnvironment) adaptableEntity);
            break;
        case ContainerrepositoryPackage.CONTAINER_TEMPLATE:
            // TODO: move to DmmModelChanger
            result = scaleInContainerTemplate((ContainerTemplate) adaptableEntity);
            break;
        default:
            throw new IllegalArgumentException("No implemented scale in operation for " + adaptableEntity);
        }
        return result;
    }

    private RuntimeEnvironment scaleOut(Entity adaptableEntity) {
        RuntimeEnvironment re = null;
        switch (adaptableEntity.eClass().getClassifierID()) {
        case ResourcelandscapePackage.RUNTIME_ENVIRONMENT:
            re = DmmModelChanger.createNewRuntimeEnvironmentInstance((RuntimeEnvironment) adaptableEntity);
            break;
        case ContainerrepositoryPackage.CONTAINER_TEMPLATE:
            re = scaleOutContainerTemplate((ContainerTemplate) adaptableEntity);
            break;
        default:
            throw new IllegalArgumentException("No implemented scale out operation for " + adaptableEntity);
        }
        return re;
    }

    private Entity scaleInContainerTemplate(ContainerTemplate template) {
        Container referringContainer = template.getReferringContainers().get(0);
        RuntimeEnvironment removedRe = null;
        if (referringContainer != null && referringContainer instanceof RuntimeEnvironment) {
            removedRe = DmmModelChanger.scaleInRuntimeEnvironmentInstance((RuntimeEnvironment) referringContainer);
        } else {
            logger.error("Cannot scale " + referringContainer);
        }

        if (removedRe != null)
            template.getReferringContainers().remove(removedRe);

        return removedRe;
    }

    /**
     * Creates a new RuntimeEnvironment. If you want to scale an attribute of the ContainerTemplate,
     * ModelVariableConfigurationRange must be used. However, here we are scaling a
     * ModelEntityConfigurationRange.
     */
    // this is just a helper function
    private RuntimeEnvironment scaleOutContainerTemplate(ContainerTemplate template) {
        if (template.getReferringContainers().size() == 0) {
            logger.debug("No referring containers found for template " + template.getName()
                    + ", creating new instance from template.");
            return DmmModelChanger.createNewRuntimeEnvironmentInstanceFromTemplate(template);
        } else {
            Container referringContainer = template.getReferringContainers().get(0);

            if (referringContainer != null && referringContainer instanceof RuntimeEnvironment) {
                return DmmModelChanger.createNewRuntimeEnvironmentInstance((RuntimeEnvironment) referringContainer);
            } else {
                logger.error("Cannot scale " + referringContainer);
            }
        }
        return null;
    }

    @Override
    public void persistActions() {
        saveModels();
    }

    // private boolean checkInvariants(EObject adaptableEntity, PropertyRange range) {
    // OclEvaluationHelper evalHelper = new OclEvaluationHelper();
    // boolean invariant = evalHelper.evaluate(adaptableEntity,
    // range.getMinValueConstraint().getOclString());
    //
    // if (!invariant)
    // logger.error("Invariant " + range.getMinValueConstraint().getOclString() + " violated");
    //
    // invariant &= evalHelper.evaluate(adaptableEntity,
    // range.getMaxValueConstraint().getOclString());
    //
    // if (!invariant)
    // logger.error("Invariant " + range.getMaxValueConstraint().getOclString() + " violated");
    //
    // return invariant;
    // }
}
